import sys
import pylab
import random
import numpy as np
import matplotlib.pyplot as plt
from collections import deque
from keras.layers import Dense
from keras.optimizers import Adam
from keras.models import Sequential
#from environment_dqn_v2 import Env
from environment_dqn_v5 import Env

EPISODES = 30000
sv = 6

# 카트폴 예제에서의 DQN 에이전트
class DQNAgent:
    def __init__(self, state_size, action_size):
        self.load_model = False

        # 상태와 행동의 크기 정의
        self.state_size = state_size
        self.action_size = action_size

        # DQN 하이퍼파라미터
        self.discount_factor = 0.99
        self.learning_rate = 0.001
        self.epsilon = 1.0
        self.epsilon_decay = 0.9995
        self.epsilon_min = 0.01
        self.batch_size = 64
        self.train_start = 10000

        # 리플레이 메모리, 최대 크기 500000
        self.memory = deque(maxlen=20000)

        # 모델과 타깃 모델 생성
        self.model = self.build_model()
        self.target_model = self.build_model()

        # 타깃 모델 초기화
        self.update_target_model()

        if self.load_model:
            self.model.load_weights("./save_model/zrm_dqn.h5")

    # 상태가 입력, 큐함수가 출력인 인공신경망 생성
    def build_model(self):
        model = Sequential()
        model.add(Dense(64, input_dim=self.state_size, activation='relu',
                        kernel_initializer='he_uniform'))
        model.add(Dense(64, activation='relu',
                        kernel_initializer='he_uniform'))
        model.add(Dense(32, activation='relu',
                        kernel_initializer='he_uniform'))
        model.add(Dense(self.action_size, activation='linear',
                        kernel_initializer='he_uniform'))
        model.summary()
        model.compile(loss='mse', optimizer=Adam(lr=self.learning_rate))
        return model

    # 타깃 모델을 모델의 가중치로 업데이트
    def update_target_model(self):
        self.target_model.set_weights(self.model.get_weights())

    # 입실론 탐욕 정책으로 행동 선택
    def get_action(self, state):
        if np.random.rand() <= self.epsilon:
            return random.randrange(self.action_size)
        else:
            q_value = self.model.predict(state)
            return np.argmax(q_value[0])

    # 샘플 <s, a, r, s'>을 리플레이 메모리에 저장
    def append_sample(self, state, action, reward, next_state, done):
        self.memory.append((state, action, reward, next_state, done))

    # 리플레이 메모리에서 무작위로 추출한 배치로 모델 학습
    def train_model(self):
        if self.epsilon > self.epsilon_min:
            self.epsilon *= self.epsilon_decay

        # 메모리에서 배치 크기만큼 무작위로 샘플 추출
        mini_batch = random.sample(self.memory, self.batch_size)

        states = np.zeros((self.batch_size, self.state_size))
        next_states = np.zeros((self.batch_size, self.state_size))
        actions, rewards, dones = [], [], []

        for i in range(self.batch_size):
            states[i] = mini_batch[i][0]
            actions.append(mini_batch[i][1])
            rewards.append(mini_batch[i][2])
            next_states[i] = mini_batch[i][3]
            dones.append(mini_batch[i][4])

        # 현재 상태에 대한 모델의 큐함수
        # 다음 상태에 대한 타깃 모델의 큐함수
        target = self.model.predict(states)
        target_val = self.target_model.predict(next_states)

        # 벨만 최적 방정식을 이용한 업데이트 타깃
        for i in range(self.batch_size):
            if dones[i]:
                target[i][actions[i]] = rewards[i]
            else:
                target[i][actions[i]] = rewards[i] + self.discount_factor * (
                    np.amax(target_val[i]))

        self.model.fit(states, target, batch_size=self.batch_size,
                       epochs=1, verbose=0)


if __name__ == "__main__":
    # CartPole-v1 환경, 최대 타임스텝 수가 500
    #env = gym.make('CartPole-v1')
    #env = wrappers.Monitor(env, '/tmp/cartpole-experiment-1', force=True)
    env = Env()
    state_size = sv + 9
    action_size = env.n_actions

    # DQN 에이전트 생성
    agent = DQNAgent(state_size, action_size)

    scores, episodes, y_mins, avg_ymins, avg_scores = [], [], [], [], []

    for e in range(EPISODES):
        done = False
        score = 0
        # env 초기화
        state = env.reset()
        state = np.reshape(state, [1, state_size])
        # y_min = max(state[0][0:37])

        while not done:
            # 현재 상태로 행동을 선택
            action = agent.get_action(state)

            # 선택한 행동으로 환경에서 한 타임스텝 진행
            next_state, reward, done = env.step(action)

            next_state = np.reshape(next_state, [1, state_size])

            # 리플레이 메모리에 샘플 <s, a, r, s'> 저장
            agent.append_sample(state, action, reward, next_state, done)

            # 매 타임스텝마다 학습
            if len(agent.memory) >= agent.train_start:
                agent.train_model()

            score += reward
            state = next_state

            if done:
                # 각 에피소드마다 타깃 모델을 모델의 가중치로 업데이트
                agent.update_target_model()

                # 에피소드마다 학습 결과 출력
                scores.append(score)
                y_mins.append(env.min_y)

                if e % 20 == 0:
                    episodes.append(e)
                    avg_ymins.append(np.mean(y_mins))
                    avg_scores.append(np.mean(scores))
                    y_mins = []
                    scores = []

                if e % 50 == 0:
                    pylab.figure(1)
                    pylab.plot(episodes, avg_ymins, 'b')
                    pylab.savefig("./save_graph/zrm_dqn_ymins.png")
                    pylab.figure(2)
                    pylab.plot(episodes, avg_scores, 'r')
                    pylab.savefig("./save_graph/zrm_dqn_scores.png")
                    print("episode:", e, "  score:", score, " min(y_max):", env.min_y, "  memory length:",
                          len(agent.memory), "  epsilon:", agent.epsilon, " steps:", env.steps)

                # 이전 20개 에피소드의 점수 평균이 490보다 크면 학습 중단
                if np.mean(avg_ymins[-min(20, len(scores)):]) < 15:
                # if e >= 499:
                    agent.model.save_weights("./save_model/zrm_dqn.h5")
                    sys.exit()

    # 학습된 제어기 실행
    Tf = 100

    shapes = np.zeros([Tf * 10, sv + 9], dtype='float32')
    inputs = np.zeros([Tf * 10, 1], dtype='int')
    state = env.reset()
    state = np.reshape(state, [1, sv + 9])

    for k in range(Tf * 10):
        action = agent.get_action(state)

        inputs[k] = action
        next_state, reward, done = env.step(action)
        shapes[k] = next_state
        next_state = np.reshape(next_state, [1, sv + 9])

        state = next_state

    plt.subplot(2, 1, 1)
    plt.plot(inputs)
    plt.subplot(2, 1, 2)
    plt.plot(shapes)
    plt.show()